#!/usr/bin/perl
#
# Jar-Tools core functions.
#
# (C) 2013,2022  ITEG IT-Engineers GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Specs:
# http://docs.oracle.com/javase/1.3/docs/guide/jar/jar.html

use strict;

use Getopt::Std;

use Cwd        qw( abs_path cwd );
use File::Path qw( make_path remove_tree );


#use Date::Format;
#use Time::Local;


# inits some global stuff

sub initFunctions() {

  $main::version = "0.9.7-2";
  $main::versionMessage = "Clazzes.org Jar-Tools ".$main::scriptName." ".$main::version."\n";

  $main::helpMessage = "jt-what?";

  $Getopt::Std::STANDARD_HELP_VERSION = 1;

  $main::commonOptionList="b:df:j:mpqv";
  $main::commonOptionsHelpText = '
 ObjectSpec := -b bundleDirectory | -j jarFile | -f manifestFile

 ObjectSpec Options:
  -b bundleDirectory   root directory of bundle, defaults to .
  -j jarFile           .jar file to investigate
  -f manifestFile      MANIFEST.MF to investigate (when -p not specified)
                       pom.properties to investigate (if -p specified)

 Harmless read-only commands define the ObjectSpec as optional
 and default to use the current directory as bundle directory.
 
 Common behaviour Options:
  -d          Debug mode, logs everything.
  -m          Parse the MANIFEST.MF, nothing else.
  -p          Parse pom.properties only, nothing else.
  -q          Quiet mode, supresses informative messages.
  -v          Be verbose, logs more.

 Common Help Options:
  --help      print this help message and exit
  --version   print version info and exit
';

}


# inits common options, i.e. default values

# call before getopts()

sub initCommonOptions() {

  %main::opts = ();

  $main::verbosity = 1;

  $main::processManifest = 1;
  $main::processPomFiles = 1; # 0 or 1?, commands may overwrite this
  $main::needPomProperties = 0;

  $main::bundleDir = ""; # or "."?
  $main::jarFile = "";
  $main::manifestFile = "";

  # GLOBALs

  $main::tempDir = "";
  $main::jarDir = "";

  # complete content
  $main::manifestContent = "";

  # sections
  @main::manifestSections = ();
  $main::nrManifestSections = 0;

  # main section details
  # name/value pairs
  @main::manifestNames = ();
  %main::manifestValues = (); # by name
  # exported packages
  @main::manifestExportedPackages = ();
  %main::manifestExportedPackageVersions = (); # by package name
  %main::manifestExportedPackageProperties = (); # by package name
  # exported services
  @main::manifestExportedServices = ();
  %main::manifestExportedServiceVersions = (); # by pure service name
  %main::manifestExportedServiceProperties = (); # by pure service name
  # imported packages
  @main::manifestImportedPackages = ();
  %main::manifestImportedPackageVersions = (); # by package name
  %main::manifestImportedPackageProperties = (); # by package name
  # imported services
  @main::manifestImportedServices = ();
  %main::manifestImportedServiceVersions = (); # by pure service name
  %main::manifestImportedServiceProperties = (); # by pure service name
  # Dojo packages
  @main::manifestDojoPackages = ();
  %main::manifestDojoPackageVersions = (); # by package name
  %main::manifestDojoPackageProperties = (); # by package name

  # stuff from pom.properties
  $main::pomPropertiesFile = "";
  @main::pomPropertiesNames = ();
  %main::pomPropertiesValues = (); # by name

}


# parses common options

# call after getopts()

sub parseCommonOptions() {

  # parse verbosity options

  if ( $main::opts{"q"} ) {
    $main::verbosity = 0;
  }

  if ( $main::opts{"d"} ) {
    if ($main::verbosity < 3) {
      $main::verbosity = 3;
    }
  }

  if ( $main::opts{"v"} ) {
    $main::verbosity++;
  }

  # parse common early options

  if ( $main::opts{"m"} ) {
    $main::processManifest = 1;
    $main::processPomFiles = 0;
  }

  if ( $main::opts{"p"} ) {
    $main::processPomFiles = 1;
    $main::needPomProperties = 1;
    $main::processManifest = 0;
  }

  # parse common other options

  my $nrObjectSpecs = 0;

  if ( defined $main::opts{"b"} ) {
    $main::bundleDir = $main::opts{"b"};
    $nrObjectSpecs++;
  }

  if ( defined $main::opts{"f"} ) {
    $nrObjectSpecs++;
    if ($main::needPomFiles) {
      $main::pomPropertiesFile = $main::opts{"f"};
    } else {
      $main::manifestFile = $main::opts{"f"};
    }
  }

  if ( defined $main::opts{"j"} ) {
    $main::jarFile = $main::opts{"j"};
    if ($main::jarFile !~ m%^/%) {
      $main::jarFile = abs_path($main::opts{"j"});
      if ($main::verbosity gt 2) { print "Absoluted jar file from ".$main::opts{"j"}." to ".$main::jarFile."\n"; }
    }
    $nrObjectSpecs++;
  }


  # check options


  if ( $nrObjectSpecs lt 1 ) {
    errorExit("FATAL: No ObjectSpec specified.");
  }
  
  if ( $nrObjectSpecs gt 1 ) {
    errorExit("FATAL: Too many ObjectSpecs specified.");
  }

  my $pomError = "";

  if ( $main::bundleDir ne "" ) {
    if ($main::verbosity gt 2) { print "Checking bundleDir ".$main::bundleDir."\n"; }
    if (! -d $main::bundleDir) {
      errorExit("FATAL: Specifed bundle dir does not exist: ".$main::bundleDir);
    }
    $main::manifestFile = $main::bundleDir."/META-INF/MANIFEST.MF";
    if (! -f $main::manifestFile) {
      errorExit("FATAL: Manifest file not in specifed bundle dir: ".$main::manifestFile);
    }
    # pom files, if of interest
    $main::pomPropertiesFile = "";
    if ($main::processPomFiles) {
      my @pomPropsInDir = glob("META-INF/maven/*/*/pom.properties");
      my $nrPomPropsInDir = scalar(@pomPropsInDir);
      if ($nrPomPropsInDir eq 1) {
        $main::pomPropertiesFile = $main::bundleDir."/".$pomPropsInDir[0];
      } else {
        $pomError .= "No or multiple ".$main::bundleDir."/META-INF/maven/*/*/pom.properties file(s).\n";
      }
    } # end if processPomFiles
  } # end if bundleDir specified

  if ( $main::jarFile ne "" ) {
    if ($main::verbosity gt 2) { print "Checking jarFile ".$main::jarFile."\n"; }
    if (! -f $main::jarFile) {
      errorExit("FATAL: Specifed jarFile file does not exist: ".$main::jarFile);
    }
    # temp dir
    eraseTempDir();
    ensureTempDir();
    $main::jarDir = $main::tempDir."/jar";
    ensureDir( $main::jarDir );
    # test unzip
    my $unzipOut = $main::tempDir."/unzip_jar_-t.out";
    if ($main::verbosity gt 0) { print "Test-unzipping ".$main::jarFile." logging to ".$unzipOut."\n"; }
    my $unzipCall = "unzip -t ".$main::jarFile." >".$unzipOut;
    my $unzipRc = system($unzipCall) >> 8;
    if ( $unzipRc ne 0 ) {
      errorExit("FATAL: Unable to   unzip -t ".$main::jarFile." >".$unzipOut);
    }
    # pom files, if of interest
    $main::pomPropertiesFile = "";
    my $pomPropertiesFileInJar = "";
    if ($main::processPomFiles) {
      # search unzip output for pom.properties files
      my $unzipLog = readCompleteFile($unzipOut);
      my @unzipLogLines = split(/\n/, $unzipLog);
      my @pomPropsInJar = ();
      foreach my $unzipLogLine (@unzipLogLines) {
        if ($unzipLogLine =~ m%\s+(META-INF/maven/[\w\.-]+/[\w\.-]+/pom.properties)[\s\$]%) {
          push @pomPropsInJar, $1;
          print "Found pom.properties in jar: ".$1."\n";
        }
      }
      my $nrPomPropsInJar = scalar(@pomPropsInJar);
      $main::pomPropertiesFile = "";
      if ($nrPomPropsInJar eq 1) {
        $pomPropertiesFileInJar = $pomPropsInJar[0];
        $main::pomPropertiesFile = $main::jarDir."/".$pomPropertiesFileInJar;
      } else {
        $pomError .= "No or multiple /META-INF/maven/*/*/pom.properties files in ".$main::jarFile."\n";
      }
    } # end if processPomFiles
    # real unzip
    if ($main::verbosity gt 0) { print "Unzipping needed files from ".$main::jarFile." to ".$main::jarDir."/\n"; }
    $unzipCall = "unzip -d ".$main::jarDir." ".$main::jarFile." META-INF/MANIFEST.MF ".$pomPropertiesFileInJar; # >/dev/null;
    $unzipRc = system($unzipCall) >> 8;
    if ( $unzipRc ne 0 ) {
      errorExit("FATAL: Unable to   unzip -t ".$main::jarFile." to ".$main::jarDir."/");
    }
    $main::manifestFile = $main::jarDir."/META-INF/MANIFEST.MF";
    if (! -f $main::manifestFile) {
      errorExit("FATAL: Jar file ".$main::jarFile." does not contain a META-INF/MANIFEST.MF file.");
    }
  }


  # check emptyness of file name properties
  if ($main::processManifest) {
    if ($main::manifestFile eq "") {
      errorExit("FATAL: Don't know what MANIFEST.MF to process");
    }
  }
  if ($main::needPomProperties) {
    if ($main::pomPropertiesFile eq "") {
      if (length($pomError) gt 0) {
        errorExit("FATAL: Don't know what pom.properties file to process.\nReason: ".$pomError);
      } else {
        errorExit("FATAL: Don't know what pom.properties file to process.");
      }
    }
  } else { 
    if ($main::pomPropertiesFile eq "") {
      $main::processPomFiles = 0;
    }
  }


  # debug infos
  if ( $main::verbosity > 2 ) {
    print "\n";
    print "Parsed common options (and configuration):\n";
    print "  Verbosity: ".$main::verbosity."\n";
    print "\n";
    print "Files to parse:\n";
    if ($main::processManifest) {
      print "  Manifest file: ".$main::manifestFile."\n";
    } else {
      print "  Manifest parsing switched off.\n";
    }
    if ($main::processPomFiles) {
      print "  pom.properties file: ".$main::pomPropertiesFile."\n";
    } else {
      print "  pom.properties parsing switched off or none present to parse.\n";
    }
    print "\n";
  }

}


# functions


# makes sure $main::tempDir is set and writeable
sub ensureTempDir() {
  if ( $main::tempDir eq "" ) {
    $main::tempDir = "/tmp/$$";
    if ($main::verbosity gt 2) { print "Checking tempDir ".$main::tempDir."\n"; }
    ensureDir($main::tempDir);
  }
}

# cleans tempDir if set and existent
sub eraseTempDir() {
  if ( $main::tempDir eq "" ) {
    return;
  }
  if ( -d $main::tempDir ) {
    if ($main::verbosity gt 2) { print "Erasing tempDir ".$main::tempDir."\n"; }
    eraseDir($main::tempDir);
  }
}


# $1 = info to print before telling about syntax and exiting
sub errorExit ( $ ) {
  my $info = shift;
  print "\n\nERROR: ".$info."\n\n";
  #print "Call   ".$main::scriptname." --help   for syntax information\n\n";
  eraseTempDir();
  exit 1;
}


# narrows header/value line (without line breaks) to not use more than 72 characters per line
# $1 = "Header: Value...."
sub narrowManifestNameValue ( $ ) {
  my $headerNameValue = shift;
  my $cpl = 72;
  my $len = length($headerNameValue);
  #if ($main::verbosity gt 2) { 
  #  print "unnarrowed = '".$headerNameValue."'\n";
  #  print "cpl=".$cpl.", len=".$len."\n";
  #}
  if ($len <= $cpl) {
    #if ($main::verbosity gt 2) { print "len ".$len." le cpl ".$cpl.", returning original value.\n"; }
    return $headerNameValue;
  }
  # ok, long version
  my $narrowed = substr($headerNameValue, 0, $cpl);
  my $l = $cpl;
  my $restLen = $len - $l;
  #if ($main::verbosity gt 2) { 
  #  print "narrowed first line: '".$narrowed."'\n";
  #  print "l = ".$l.", restLen = ".$restLen."\n";
  #}
  while ($restLen > 0) {
    if ($restLen >= $cpl) {
      $narrowed .= "\n ".substr($headerNameValue, $l, ($cpl-1));
      $l += ($cpl-1);
      $restLen = $len - $l;
      #if ($main::verbosity gt 2) { print "new l = ".$l.", restLen = ".$restLen."\n"; }
    } else {
      $narrowed .= "\n ".substr($headerNameValue, $l);
      $restLen = 0;
    }
  }
  #if ($main::verbosity gt 2) { print "Final narrowed: '".$narrowed."'\n"; }
  return $narrowed;
}



# read MANIFEST.MF
sub readManifest () {

  if ( $main::manifestFile eq "" ) {
    errorExit("FATAL: readManifest called but no manifestFile set!");
  }
  # read it
  if ($main::verbosity gt 1) { print "Reading manifest from ".$main::manifestFile."\n"; }
  $main::manifestContent = readCompleteFile($main::manifestFile);
  if ($main::verbosity gt 2) { print "Read ".length($main::manifestContent)." chars from ".$main::manifestFile."\n"; }

  # first unify linebreaks
  $main::manifestContent =~ s/\r\n/\n/g;
  $main::manifestContent =~ s/\r/\n/g;

  # section split
  @main::manifestSections = split(/\n\n/, $main::manifestContent);
  $main::nrManifestSections = scalar(@main::manifestSections);
  if ($main::verbosity gt 3) { 
    print "Manifest has ".$main::nrManifestSections." sections.\n\n";
    print "Main section:\n".$main::manifestSections[0]."---\n\n";
    if ($main::nrManifestSections gt 1) {
      for (my $i = 1; $i < $main::nrManifestSections; $i++) {
        print "Further section ".$i.":\n".$main::manifestSections[$i]."---\n\n";
      }
    } else {
      print "No further sections.\n\n";
    }
  }

  my @lines = split(/\n/, $main::manifestSections[0]);
  my $lastName = "";
  my $lineNr = 0;
  @main::manifestNames = ();
  %main::manifestValues = (); # by name
  foreach my $line (@lines) {
    $lineNr++;
    if ($line =~ m/^#/) { next; }
    if ($line =~ m/^ /) {
      if (length($lastName) > 0) {
        $main::manifestValues{$lastName} .= substr($line, 1);
      } else {
        print "Ignoring illegal continuation line at ".$main::manifestFile.":".$lineNr."\n";
        $lastName = "";
      }
      next;
    }
    if (not($line =~ /[\w\.-].*: .+/)) {
      print "Ignoring illegal name/value line at ".$main::manifestFile.":".$lineNr."\n";
      next;
    }
    my ($name, $value) = split(/: /, $line, 2);
    # evtl. check for repeated header line
    push @main::manifestNames, $name;
    $main::manifestValues{$name} = $value;
    $lastName = $name;
  }
  if ($main::verbosity gt 3) { 
    print "Manifest's main section has these name/value pairs:\n\n";
    foreach my $name (@main::manifestNames) {
      print "Name: ".$name."\n".
            "Value: ".$main::manifestValues{$name}."\n\n";
    }
  }
  # parse exported packages
  @main::manifestExportedPackages = ();
  %main::manifestExportedPackageVersions = (); # by pure package name
  %main::manifestExportedPackageProperties = (); # by package name
  my $exportedPackagesList = $main::manifestValues{"Export-Package"};
  if (length($exportedPackagesList)) {
    if ($main::verbosity gt 1) { print "Analysing value of Export-Package.\n"; }
    @main::manifestExportedPackages = splitPackageList($exportedPackagesList);
    foreach my $packageSpec (@main::manifestExportedPackages) {
      my ($packageName, $packageProperties, $packageVersion) = splitPackageSpec($packageSpec);
      $main::manifestExportedPackageVersions{$packageName} = $packageVersion;
      $main::manifestExportedPackageProperties{$packageName} = $packageProperties;
    }
  } else {
    if ($main::verbosity gt 1) { print "No Export-Package value to analyse.\n"; }
  }
  # parse exported services
  @main::manifestExportedServices = ();
  %main::manifestExportedServiceVersions = (); # by pure service name
  %main::manifestExportedServiceProperties = (); # by pure service name
  my $exportedServicesList = $main::manifestValues{"Export-Service"};
  if (length($exportedServicesList)) {
    if ($main::verbosity gt 1) { print "Analysing value of Export-Service.\n"; }
    @main::manifestExportedServices = splitPackageList($exportedServicesList);
    foreach my $serviceSpec (@main::manifestExportedServices) {
      my ($serviceName, $serviceProperties, $serviceVersion) = splitPackageSpec($serviceSpec);
      $main::manifestExportedServiceVersions{$serviceName} = $serviceVersion;
      $main::manifestExportedServiceProperties{$serviceName} = $serviceProperties;
    }
  } else {
    if ($main::verbosity gt 1) { print "No Export-Service value to analyse.\n"; }
  }
  # parse imported packages
  @main::manifestImportedPackages = ();
  %main::manifestImportedPackageVersions = (); # by pure package name
  %main::manifestImportedPackageProperties = (); # by package name
  my $importedPackagesList = $main::manifestValues{"Import-Package"};
  if (length($importedPackagesList)) {
    if ($main::verbosity gt 1) { print "Analysing value of Import-Package.\n"; }
    @main::manifestImportedPackages = splitPackageList($importedPackagesList);
    foreach my $packageSpec (@main::manifestImportedPackages) {
      my ($packageName, $packageProperties, $packageVersion) = splitPackageSpec($packageSpec);
      $main::manifestImportedPackageVersions{$packageName} = $packageVersion;
      $main::manifestImportedPackageProperties{$packageName} = $packageProperties;
    }
  } else {
    if ($main::verbosity gt 1) { print "No Import-Package value to analyse.\n"; }
  }
  # parse imported services
  @main::manifestImportedServices = ();
  %main::manifestImportedServiceVersions = (); # by pure service name
  %main::manifestImportedServiceProperties = (); # by pure service name
  my $importedServicesList = $main::manifestValues{"Import-Service"};
  if (length($importedServicesList)) {
    if ($main::verbosity gt 1) { print "Analysing value of Import-Service.\n"; }
    @main::manifestImportedServices = splitPackageList($importedServicesList);
    foreach my $serviceSpec (@main::manifestImportedServices) {
      my ($serviceName, $serviceProperties, $serviceVersion) = splitPackageSpec($serviceSpec);
      $main::manifestImportedServiceVersions{$serviceName} = $serviceVersion;
      $main::manifestImportedServiceProperties{$serviceName} = $serviceProperties;
    }
  } else {
    if ($main::verbosity gt 1) { print "No Import-Service value to analyse.\n"; }
  }
  # parse dojo packages
  @main::manifestDojoPackages = ();
  %main::manifestDojoPackageVersions = (); # by pure package name
  %main::manifestDojoPackageProperties = (); # by package name
  my $dojoPackagesList = $main::manifestValues{"X-Dojo-Package"};
  if (length($dojoPackagesList)) {
    if ($main::verbosity gt 1) { print "Analysing value of X-Dojo-Package.\n"; }
    @main::manifestDojoPackages = splitPackageList($dojoPackagesList);
    foreach my $packageSpec (@main::manifestDojoPackages) {
      my ($packageName, $packageProperties, $packageVersion) = splitPackageSpec($packageSpec);
      $main::manifestDojoPackageVersions{$packageName} = $packageVersion;
      $main::manifestDojoPackageProperties{$packageName} = $packageProperties;
    }
  } else {
    if ($main::verbosity gt 2) { print "No X-Dojo-Package value to analyse.\n"; }
  }

}
# end readManifest


# read pom.properties
sub readPomFiles () {

  if ( $main::pomPropertiesFile eq "" ) {
    errorExit("FATAL: readPomFiles called but no pomPropertiesFile set!");
  }
  # read it
  if ($main::verbosity gt 1) { print "Reading pom files ".$main::pomPropertiesFile."\n"; }
  $main::pomPropertiesContent = readCompleteFile($main::pomPropertiesFile);
  if ($main::verbosity gt 2) { print "Read ".length($main::pomPropertiesContent)." chars from ".$main::pomPropertiesFile."\n"; }

  # first unify linebreaks
  $main::pomPropertiesContent =~ s/\r\n/\n/g;
  $main::pomPropertiesContent =~ s/\r/\n/g;
  # lines split
  my @lines = split(/\n/, $main::pomPropertiesContent);
  @main::pomPropertiesNames = ();
  %main::pomPropertiesValues = (); # by name
  my $lineNr = 0;
  foreach my $line (@lines) {
    $lineNr++;
    if (length($line) eq 0) { next; }
    if ($line =~ m/^#/) { next; }
    if (not($line =~ /[a-zA-Z].*=.+/)) {
      print "Ignoring illegal name/value line at ".$main::pomPropertiesFile.":".$lineNr."\n";
      next;
    }
    my ($name, $value) = split(/=/, $line, 2);
    push @main::pomPropertiesNames, $name;
    $main::pomPropertiesValues{$name} = $value;
  }
  if ($main::verbosity gt 3) { 
    print "pom.properties has these name/value pairs:\n\n";
    foreach my $name (@main::pomPropertiesNames) {
      print "Name: ".$name."\n".
            "Value: ".$main::pomPropertiesValues{$name}."\n\n";
    }
  }
}
# end readPomFiles


sub printManifestMetaInfos() {

  print "Meta infos from manifest:\n\n";

  if (length($main::manifestValues{"Bundle-Name"}) gt 0 ) {
    print "Bundle-* attributes:\n";
    print "  Name: ".$main::manifestValues{"Bundle-Name"}."\n";
    print "  Symbolic name: ".$main::manifestValues{"Bundle-SymbolicName"}."\n";
    print "  Version: ".$main::manifestValues{"Bundle-Version"}."\n";
    print "  Vendor: ".$main::manifestValues{"Bundle-Vendor"}."\n";
    print "  Required environment: ".$main::manifestValues{"Bundle-RequiredExecutionEnvironment"}."\n";
    print "  Description: ".$main::manifestValues{"Bundle-Description"}."\n";
  } else {
    print "No Bundle-Name found, not search for other Bundle-* attributes:\n";
  }
  print "\n";

  my $nrExportedPackages = scalar(@main::manifestExportedPackages);
  if ($nrExportedPackages eq 0) {
    print "No packages exported.\n";
  } else {
    print "".$nrExportedPackages." packages exported:\n";
    foreach my $packageName (sort(keys(%main::manifestExportedPackageVersions))) {
      my $packageVersion = $main::manifestExportedPackageVersions{$packageName};
      my $packageProperties = $main::manifestExportedPackageProperties{$packageName};
      printPackageInfos($packageName, $packageVersion, $packageProperties);
    }
  }
  print "\n";
    
  my $nrExportedServices = scalar(@main::manifestExportedServices);
  if ($nrExportedServices eq 0) {
    print "No services exported.\n";
  } else {
    print "".$nrExportedServices." services exported:\n";
    foreach my $serviceName (sort(keys(%main::manifestExportedServiceVersions))) {
      my $serviceVersion = $main::manifestExportedServiceVersions{$serviceName};
      my $serviceProperties = $main::manifestExportedServiceProperties{$serviceName};
      printPackageInfos($serviceName, $serviceVersion, $serviceProperties);
    }
  }
  print "\n";

  my $nrImportedPackages = scalar(@main::manifestImportedPackages);
  if ($nrImportedPackages eq 0) {
    print "No packages imported.\n";
  } else {
    print "".$nrImportedPackages." packages imported:\n";
    foreach my $packageName (sort(keys(%main::manifestImportedPackageVersions))) {
      my $packageVersion = $main::manifestImportedPackageVersions{$packageName};
      my $packageProperties = $main::manifestImportedPackageProperties{$packageName};
      printPackageInfos($packageName, $packageVersion, $packageProperties);
    }
  }
  print "\n";

  my $nrImportedServices = scalar(@main::manifestImportedServices);
  if ($nrImportedServices eq 0) {
    print "No services imported.\n";
  } else {
    print "".$nrImportedServices." services imported:\n";
    foreach my $serviceName (sort(keys(%main::manifestImportedServiceVersions))) {
      my $serviceVersion = $main::manifestImportedServiceVersions{$serviceName};
      my $serviceProperties = $main::manifestImportedServiceProperties{$serviceName};
      printPackageInfos($serviceName, $serviceVersion, $serviceProperties);
    }
  }
  print "\n";

  my $bundleClassPath = "";
  if (defined $main::manifestValues{"Bundle-ClassPath"}) {
    $bundleClassPath = $main::manifestValues{"Bundle-ClassPath"};
  }
  if (length($bundleClassPath) eq 0) {
    print "No Bundle-ClassPath specified.\n";
  } else {
    my $bundleClassPath = $main::manifestValues{"Bundle-ClassPath"};
    print "Bundle-ClassPath: ".$bundleClassPath."\n";
  }
  print "\n";
  
  my $nrDojoPackages = scalar(@main::manifestDojoPackages);
  if ($nrDojoPackages eq 0) {
    if ($main::verbosity gt 1) { 
      print "No X-Dojo-Packages header.\n";
      print "\n";
    }
  } else {
    print "".$nrDojoPackages." Dojo packages mentioned:\n";
    foreach my $packageName (sort(keys(%main::manifestDojoPackageVersions))) {
      my $packageVersion = $main::manifestDojoPackageVersions{$packageName};
      my $packageProperties = $main::manifestDojoPackageProperties{$packageName};
      printPackageInfos($packageName, $packageVersion, $packageProperties);
    }
    print "\n";
  }
    

}
# end printExportsImports()


# $1 = package name
# $2 = package version
# $3 = full package properties
sub printPackageInfos ( $ $ $ ) {
  my $packageName = shift;
  my $packageVersion = shift;
  my $packageProperties = shift;
  print "  ".$packageName;
  if (length($packageVersion) gt 0) {
    print " ".$packageVersion;
  }
  # construct expr. to remove version part from properties
  my $versionProperty = 'version="'.$packageVersion.'"';
  $versionProperty =~ s/\[/\\[/;
  $versionProperty =~ s/\]/\\]/;
  $versionProperty =~ s/\(/\\(/;
  $versionProperty =~ s/\)/\\)/;
  # remove version part from properties
  $packageProperties =~ s/$versionProperty//;
  # clear bad ;
  $packageProperties =~ s/;;/;/g;
  $packageProperties =~ s/^;//;
  $packageProperties =~ s/;$//;
  # show properties if some remaining
  if (length($packageProperties) gt 0) {
    print "\n    ".$packageProperties;
  }
  print "\n";
}
# end printPackageInfos 


sub printPomMetaInfos() {

  print "Meta infos from pom files:\n\n";

  if ($main::pomPropertiesFile ne "") {
    print "From pom.properties:\n";
    print "  Group ID: ".$main::pomPropertiesValues{"groupId"}."\n";
    print "  Artifact ID: ".$main::pomPropertiesValues{"artifactId"}."\n";
    print "  Version: ".$main::pomPropertiesValues{"version"}."\n";
  } else {
    print "No pom.properties file parsed.\n";
  }
  print "\n";
} # print pomMetaInfos


sub recollectManifestContent () {
  if ($main::verbosity gt 2) { print "Recollecting manifest content.\n"; }
  # headers
  my $newMainSection = "";
  foreach my $headerName (@main::manifestNames) {
    if ($main::verbosity gt 3) { print "Adding header ".$headerName."\n"; }
    $newMainSection .= narrowManifestNameValue($headerName.": ".$main::manifestValues{$headerName}) . "\n";
  }
  if ($main::verbosity gt 3) { 
    print "New main section:\n".$newMainSection."---\n\n";
  }
  $main::manifestSections[0] = $newMainSection;
  if ($main::nrManifestSections gt 1) {
    chomp $main::manifestSections[0];
  }
  $main::manifestContent = join("\n\n", @main::manifestSections);
} # end recollectManifestContent


# $1 = raw package list
# returns list split at non-quoted commas
sub splitPackageList ( $ ) {
  my @packageSpecs = ();
  my $packageList = shift;
  my $len = length($packageList);
  if ($len eq 0) {
    return @packageSpecs;
  }
  #my $inQuote = 0;
  my $inSingleQuotes = 0;
  my $inDoubleQuotes = 0;
  my $packageSpecBegin = 0;
  if ($main::verbosity gt 3) { 
    print "Starting to analyse this packageList:\n".$packageList."\n";
  }
  for (my $i; $i < $len; $i++) {
    my $c = substr($packageList, $i, 1);
    #print "c=$c\n";
    # first avoid any quoting confusion
    if ($inDoubleQuotes) {
      if ($c eq '"') {
        $inDoubleQuotes = 0;
        #print "Closing double quotes\n";
      }
      next;
    }
    if ($inSingleQuotes) {
      if ($c eq "'") {
        $inSingleQuotes = 0;
        #print "Closing single quotes\n";
      }
      next;
    }
    # check for opening quotes
    if ($c eq '"') {
      $inDoubleQuotes = 1;
      #print "Opening double quotes\n";
      next;
    }
    if ($c eq "'") {
      $inSingleQuotes = 1;
      #print "Opening single quotes\n";
      next;
    }
    # we are outside quotes, check for comma
    if ($c eq ',') {
      #print "Found unquoted comma\n";
      push @packageSpecs, substr($packageList, $packageSpecBegin, $i-$packageSpecBegin);
      $packageSpecBegin = $i+1;
    }
  }
  if ($packageSpecBegin < $len) {
    push @packageSpecs, substr($packageList, $packageSpecBegin);
  }
  if ($main::verbosity gt 3) {
    print "Returning these packageSpecs:\n";
    foreach my $packageSpec (@packageSpecs) {
      print "  ".$packageSpec."\n";
    }
    print "\n";
  }
  return @packageSpecs;
}
# end splitPackageList


# $1 = packageSpec as returend by splitPackageList
# returns ( $packageName, $packageProperties, $packageVersion );
sub splitPackageSpec ( $ ) {
  my $packageSpec = shift;
  my ($packageName, $packageProperties) = split(/;/, $packageSpec, 2);
  my $packageVersion = "";
  my $regex = '.*version="([^"]+)"';
  if ($packageSpec =~ m/$regex/) {
    $packageVersion = $1;
  }
  return ( $packageName, $packageProperties, $packageVersion );
}


# write MANIFEST.MF
sub writeManifest () {
  if ( $main::manifestFile eq "" ) {
    errorExit("FATAL: writeManifest called but no manifestFile set!");
  }
  # write file
  if ($main::verbosity gt 1) { print "Writing ".length($main::manifestContent)." chars to ".$main::manifestFile."\n"; }
  open (NEWMANIFEST, ">".$main::manifestFile) || exitError("Unable to write new manifest to ".$main::manifestFile);
  print NEWMANIFEST $main::manifestContent;
  close (NEWMANIFEST);
  if ($main::verbosity gt 2) { print "Done writing ".length($main::manifestContent)." chars to ".$main::manifestFile."\n"; }
  # eventually re-zip
  my $callDir = cwd();
  if ( $main::jarFile ne "" ) {
    print "Updating ".$main::jarFile." with new META-INF/MANIFEST.MF\n";
    chdir $main::jarDir || errorExit("Cannot chdir to ".$main::jarDir);
    if ($main::verbosity gt 1) { print "Updating ".$main::jarFile." with new META-INF/MANIFEST.MF\n"; }
    my $zipCall = "zip ".$main::jarFile." META-INF/MANIFEST.MF";
    if ($main::verbosity gt 2) { print "From ".$main::jarDir." calling ".$zipCall."\n"; }
    my $zipRc = system($zipCall) >> 8;
    if ( $zipRc ne 0 ) {
      errorExit("FATAL: Unable to   zip ".$main::jarFile."  META-INF/MANIFEST.MF");
    }
  } else {
    if ($main::verbosity gt 1) { print "No need to update any .jar file.\n"; }
    print "No need to update any .jar file.\n";
  }
  chdir $callDir;
}
# end writeManifest


# minor helpers

#sub dirName ( $ ) {
#  my $file = shift;
#  my $len = length($file);
#  if ($len < 2) {
#    errorExit("dirName() called with illegal value ".$file);
#  }
#  my $lastSlash = rindex($file, '/', $len-2); # -1?
#  #print "lastSlash=".$lastSlash."\n";
#  if ($lastSlash < 0 or $lastSlash > $len) {
#    errorExit("dirName() called with illegal value ".$file);
#  }
#  my $dir = substr($file, 0, $lastSlash);
#  if ($dir eq "") { $dir = "/"; }
#  return $dir;
#}

sub ensureDir( $ ) {
  my $dir = shift;
  if ( ! -d $dir ) {
    if ( $main::verbosity gt 2 ) {
      print "Auto-creating path ".$dir."\n";
      make_path($dir, { mode => 0700, verbose => 1 }) || errorExit("FATAL: Unable to create path ".$dir);
    } else {
      make_path($dir, { mode => 0700, verbose => 0 }) || errorExit("FATAL: Unable to create path ".$dir);
    }
  }
}

sub eraseDir( $ ) {
  my $dir = shift;
  if ( -d $dir ) {
    if ( $main::verbosity gt 3 ) {
      print "Removing path ".$dir."\n";
      remove_tree($dir, { verbose => 1 }) || die "FATAL: Unable to remove path ".$dir."\n";
                                                           # errorExit() might cause endless loop, so use straight die
    } else {
      remove_tree($dir, { verbose => 0 }) || die "FATAL: Unable to remove path ".$dir."\n";
                                                           # errorExit() might cause endless loop, so use straight die
    }
  }
}


# $1 = name of text file
sub readCompleteFile( $ ) {
  # returns content of file
  my $filename=shift;
  my $text;
  my $defaultLF=$/;
  undef $/;
  open (TMPFILE, "<".$filename) || errorExit("FATAL: Cannot read ".$filename);
  $text=<TMPFILE>;
  close TMPFILE;
  $/=$defaultLF;
  return $text;
}


# for GetOpt:

sub main::HELP_MESSAGE() {
  print $main::helpMessage;
}

sub main::VERSION_MESSAGE() {
  print $main::versionMessage;
}


# END block
END {
  #eraseTempDir();
}


1;

